import argparse
import logging
import requests
import sys

from crlite_query import CRLiteDB, CRLiteQuery, IntermediatesDB, parse_hosts_file
from datetime import datetime, timedelta
from pathlib import Path
from urllib.parse import urlparse

log = logging.getLogger("query_cli")


crlite_collection_prod = (
    "https://firefox.settings.services.mozilla.com/v1/buckets/security-state"
    + "/collections/cert-revocations/records"
)
crlite_collection_stage = (
    "https://settings.stage.mozaws.net/v1/buckets/security-state"
    + "/collections/cert-revocations/records"
)
intermediates_collection_prod = (
    "https://firefox.settings.services.mozilla.com/v1/buckets/security-state"
    + "/collections/intermediates/records"
)


def find_attachments_base_url(urlstring):
    url = urlparse(urlstring)
    base_rsp = requests.get(f"{url.scheme}://{url.netloc}/v1/")
    return base_rsp.json()["capabilities"]["attachments"]["base_url"]


def main():
    parser = argparse.ArgumentParser(
        description="Query CRLite data",
        epilog="""
      The --db option should point to a folder containing a single filter file of
      the form "YYYYMMDDnn.filter" along with a collection of files of the form
      "YYYYMMDDnn.stash" which contain updates from that original filter. By
      default, if this tool believes it is out-of-date based on the local
      database, it will attempt to update itself before performing its checks.
      To avoid that behavior, pass --no-update on the command line.
    """,
    )
    parser.add_argument(
        "--hosts",
        help="Hosts to check, in the form host[:port] where "
        + "port is assumed 443 if not provided. Can be specified multiple times.",
        action="append",
        nargs="+",
        default=[],
        metavar="host[:port]",
    )
    parser.add_argument(
        "--hosts-file",
        help="File of hosts to check, in the form of 'host[:port]' each line, "
        + "where port is assumed 443 if not provided. Can be specified multiple "
        + " times.",
        action="append",
        default=[],
        type=Path,
    )
    parser.add_argument(
        "files", help="PEM files to load", type=argparse.FileType("r"), nargs="*"
    )
    parser.add_argument(
        "--db",
        type=Path,
        default=Path("~/.crlite_db"),
        help="Path to CRLite database folder",
    )
    parser.add_argument(
        "--no-update", help="Do not attempt to update the database", action="store_true"
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--force-update", help="Force an update to the database", action="store_true"
    )
    group.add_argument(
        "--use-filter",
        help="Use this specific filter file, ignoring the database",
        type=Path,
    )
    parser.add_argument(
        "--check-not-revoked",
        help="Set exit code 0 if none of the supplied certificates are revoked",
        action="store_true",
    )
    parser.add_argument(
        "--no-delete",
        help="Do not attempt to delete old database files",
        action="store_true",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--crlite-url",
        default=crlite_collection_prod,
        help="URL to the CRLite records at Remote Settings.",
    )
    group.add_argument(
        "--crlite-staging",
        action="store_true",
        help="Use the staging URL for CRLite",
    )
    parser.add_argument(
        "--intermediates-url",
        default=intermediates_collection_prod,
        help="URL to the CRLite records at Remote Settings.",
    )
    group.add_argument(
        "--download-intermediates",
        action="store_true",
        help="Download all intermediate PEM files to the database",
    )
    parser.add_argument(
        "--verbose", "-v", help="Be more verbose", action="count", default=0
    )
    group.add_argument(
        "--structured",
        help="Emit log entries intended for structured loggers",
        action="store_true",
    )

    args = parser.parse_args()

    if args.crlite_staging:
        args.crlite_url = crlite_collection_stage

    if args.verbose > 0:
        logging.basicConfig(level=logging.DEBUG)
        if args.verbose > 1:
            from pyasn1 import debug

            debug.setLogger(debug.Debug("all"))
    else:
        logging.basicConfig(level=logging.INFO)

    db_dir = args.db.expanduser()

    if not db_dir.is_dir():
        db_dir.expanduser().mkdir()

    last_updated_file = (db_dir / ".last_updated").expanduser()
    if last_updated_file.exists() and not args.force_update:
        updated_file_timestamp = datetime.fromtimestamp(
            last_updated_file.stat().st_mtime
        )
        grace_time = datetime.now() - timedelta(hours=6)
        if last_updated_file.is_file() and updated_file_timestamp > grace_time:
            log.info(f"Database was updated at {updated_file_timestamp}, skipping.")
            log.debug(
                f"Database was last updated {datetime.now() - updated_file_timestamp} ago."
            )
            args.no_update = True

    attachments_base_url = find_attachments_base_url(args.crlite_url)

    intermediates_db = IntermediatesDB(
        db_path=db_dir, download_pems=args.download_intermediates
    )
    crlite_db = CRLiteDB(db_path=args.db)

    try:
        if args.force_update or not args.no_update:
            if args.download_intermediates:
                log.info(
                    "Downloading all intermediate certificates. Look in "
                    + f"{intermediates_db.intermediates_path}"
                )

            intermediates_db.update(
                collection_url=args.intermediates_url,
                attachments_base_url=attachments_base_url,
            )
            crlite_db.update(
                collection_url=args.crlite_url,
                attachments_base_url=attachments_base_url,
            )
            last_updated_file.touch()
    except KeyboardInterrupt:
        log.warning("Interrupted.")
        sys.exit(1)

    if args.use_filter:
        crlite_db.load_filter(path=args.use_filter)

    if not args.no_delete:
        crlite_db.cleanup()

    log.info(f"Status: {intermediates_db}, {crlite_db}")

    query = CRLiteQuery(intermediates_db=intermediates_db, crlite_db=crlite_db)

    if not args.files and not args.hosts and not args.hosts_file:
        log.info("No PEM files or hosts specified to load. Run with --help for usage.")

    to_test = list()

    for file in args.files:
        to_test.append((file.name, query.gen_from_pem(file)))

    host_strings = []
    for host_list in args.hosts:
        host_strings.extend(host_list)

    for path in args.hosts_file:
        with path.open("r") as fd:
            host_strings.extend(parse_hosts_file(fd))

    for host_str in host_strings:
        parts = host_str.split(":")
        hostname = parts[0]
        port = 443
        if len(parts) > 1:
            port = int(parts[1])
        to_test.append((f"{hostname}:{port}", query.gen_from_host(hostname, port)))

    failures = list()

    for (name, generator) in to_test:
        for result in query.query(name=name, generator=generator):
            if args.structured:
                result.log_query_result()
            else:
                result.print_query_result()

            if args.check_not_revoked and result.is_revoked():
                failures.append(result)

    if failures:
        print(f"{len(failures)} failures logged:")
        for result in failures:
            print(result)
        sys.exit(1)


if __name__ == "__main__":
    main()



import json
import io
import socket
import tempfile
import unittest
import crlite_query

from datetime import datetime, timezone
from pathlib import Path
from crlite_query import CRLiteDB, IntermediatesDB
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread


class MockCollectionRequestHandler(BaseHTTPRequestHandler):
    def serve_collection(self, files):
        if "/collection" in self.path:
            data = []
            for entry in files:
                collection_item = {
                    "schema": 1,
                    "attachment": {
                        "hash": entry["hash"],
                        "size": entry["path"].stat().st_size,
                        "filename": entry["path"].name,
                        "location": entry["path"].name,
                        "mimetype": entry["content-type"],
                    },
                    "id": entry["hash"],
                    "last_modified": 1,
                }
                collection_item.update(entry["extra"])
                data.append(collection_item)

            self.send_response(200)
            self.send_header("Content-Type", "application/json; charset=utf-8")
            self.end_headers()
            self.wfile.write(json.dumps({"data": data}).encode("utf-8"))
            return

        for e in files:
            if self.path == "/attachments/" + e["path"].name:
                self.send_response(200)
                self.send_header("Content-Type", e["content-type"])
                self.end_headers()
                self.wfile.write(e["path"].read_bytes())
                return

        self.send_error(404, message=self.path)
        return


class MockIntermediatesDataRequestHandler(MockCollectionRequestHandler):
    def do_GET(self):
        self.serve_collection(
            [
                {
                    "path": Path(__file__).resolve().parent
                    / Path("test-intermediate.pem"),
                    "hash": "70f02be25c991671f01c41506e2963211ca7c9b89b59ffad7c412af8b5e1f40c",
                    "content-type": "application/x-pem-file",
                    "extra": {
                        "derHash": "hash",
                        "subject": "subj",
                        "subjectDN": "ZGlzdGluZ3Vpc2hlZE5hbWU=",
                        "pubKeyHash": "hash",
                        "whitelist": False,
                        "crlite_enrolled": True,
                    },
                }
            ]
        )


class MockCRLiteDataRequestHandler(MockCollectionRequestHandler):
    def do_GET(self):
        self.serve_collection(
            [
                {
                    "path": Path(__file__).resolve().parent / Path("test-1.filter"),
                    "hash": "80e8e148fbf95aed39783f1fcc2d4576074f8c487656ca2d53571da4b17e20a9",
                    "content-type": "application/octet-stream",
                    "extra": {
                        "details": {"name": "2020-04-02T06:00:00Z-full"},
                        "incremental": False,
                        "effectiveTimestamp": 1585807200000,
                    },
                },
                {
                    "path": Path(__file__).resolve().parent / Path("test-1.stash"),
                    "hash": "bda63519578e451eeef9c828a003ee249c56b4cb97bdf2909e7563851d2bd985",
                    "content-type": "application/octet-stream",
                    "extra": {
                        "details": {"name": "2020-04-02T12:00:00Z-diff"},
                        "incremental": True,
                        "effectiveTimestamp": 1585828800000,
                        "parent": "80e8e148fbf95aed39783f1fcc2d4576074f8c487656ca2d53571da4b17e20a9",
                    },
                },
            ]
        )


class MockServer(object):
    def __init__(self, handler):
        # Configure mock server.
        self.server_port = MockServer.get_free_port()
        self.server = HTTPServer(("localhost", self.server_port), handler)
        self.base_uri = f"http://localhost:{self.server_port}"

    @classmethod
    def get_free_port(cls):
        s = socket.socket(socket.AF_INET, type=socket.SOCK_STREAM)
        s.bind(("localhost", 0))
        address, port = s.getsockname()
        s.close()
        return port

    def __enter__(self):
        # Start running mock server in a separate thread.
        # Daemon threads automatically shut down when the main process exits.
        self.server_thread = Thread(target=self.server.serve_forever)
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        return self.base_uri

    def __exit__(self, type, value, tb):
        self.server.shutdown()
        self.server.server_close()


class TestCRLiteDB(unittest.TestCase):
    def test_load_empty_dir(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            db = CRLiteDB(db_path=temp_dir)

            self.assertEqual(db.filter_file, None)
            self.assertEqual(db.stash_files, [])
            self.assertEqual(db.latest_covered_date(), None)

            with MockServer(MockCRLiteDataRequestHandler) as base_uri:
                db.update(
                    collection_url=f"{base_uri}/collection/",
                    attachments_base_url=f"{base_uri}/attachments/",
                )

            self.assertNotEqual(db.filter_file, None)
            self.assertEqual(len(db.stash_files), 1)
            self.assertEqual(
                db.latest_covered_date(),
                datetime(2020, 4, 2, 12, 0, tzinfo=timezone.utc),
            )

    def test_load_explicit_filter(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            db = CRLiteDB(db_path=temp_dir)

            self.assertEqual(db.filter_file, None)
            self.assertEqual(db.stash_files, [])
            self.assertEqual(db.latest_covered_date(), None)

            db.load_filter(path=Path(__file__).resolve().parent / Path("test-1.filter"))
            self.assertEqual(db.stash_files, [])

            # Since the filter's name doesn't include the datestamp, expect
            # a warning.
            with self.assertLogs(level="WARNING") as log:
                self.assertNotEqual(db.latest_covered_date(), None)
                self.assertIn("Invalid isoformat string", log.output[0])


class TestIntermediatesDB(unittest.TestCase):
    def test_load_empty_dir(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            db = IntermediatesDB(db_path=temp_dir, download_pems=True)
            self.assertTrue(db.intermediates_path.exists())
            self.assertEqual(list(db.intermediates_path.iterdir()), [])

            with MockServer(MockIntermediatesDataRequestHandler) as base_uri:
                db.update(
                    collection_url=f"{base_uri}/collection/",
                    attachments_base_url=f"{base_uri}/attachments/",
                )

            self.assertEqual(len(list(db.intermediates_path.iterdir())), 1)

            self.assertEqual(len(db), 1)
            self.assertTrue("1 Intermediate" in str(db))

            issuer = db.issuer_by_DN(b"unknownName")
            self.assertEqual(issuer, None)

            issuer = db.issuer_by_DN(b"distinguishedName")
            self.assertNotEqual(issuer, None)
            self.assertEqual(issuer["crlite_enrolled"], True)
            self.assertTrue(issuer["path"].exists())


class TestHostFileParsing(unittest.TestCase):
    def test_empty_file(self):
        results = crlite_query.parse_hosts_file(io.StringIO(""))
        self.assertFalse(results)

    def test_only_comments(self):
        results = crlite_query.parse_hosts_file(
            io.StringIO(
                """
            # this is a comment
            ; and so is this

            # and that was a blank line
        """
            )
        )
        self.assertFalse(results)

    def test_one_host_and_comments(self):
        results = crlite_query.parse_hosts_file(
            io.StringIO(
                """
            # this is a comment
            ; and so is this
            example.com:999
            # and that was a blank line
        """
            )
        )
        self.assertEqual(results, ["example.com:999"])

    def test_two_hosts_and_comments(self):
        results = crlite_query.parse_hosts_file(
            io.StringIO(
                """
            example.net
            \t# this is a comment
            ; \tand so is this
            example.com:999
            # okay we should be done now
        """
            )
        )
        self.assertEqual(results, ["example.net", "example.com:999"])

    def test_one_host(self):
        results = crlite_query.parse_hosts_file(
            io.StringIO(
                """
            example.com:112
        """
            )
        )
        self.assertEqual(results, ["example.com:112"])


if __name__ == "__main__":
    unittest.main()



import base64
import collections
import hashlib
import logging
import progressbar
import re
import requests
import socket
import sqlite3
import ssl
import sys

from datetime import datetime, timezone
from filtercascade import FilterCascade
from moz_crlite_lib import CertId, IssuerId, readFromAdditionsList
from pathlib import Path
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.type import namedtype
from pyasn1.type import tag
from pyasn1.type import univ
from pyasn1_modules import pem
from pyasn1_modules import rfc2459
from urllib.parse import urljoin

log = logging.getLogger("crlite_query")

assert sqlite3.sqlite_version_info >= (3, 24), "Requires SQLite 3.24 or newer"
assert sys.version_info >= (3, 7), "Requires Python 3.7 or newer"


def ensure_local(*, base_url, entry, local_path):
    url = urljoin(base_url, entry["attachment"]["location"])
    if local_path.is_file():
        h = hashlib.sha256()
        h.update(local_path.read_bytes())
        if h.hexdigest() == entry["attachment"]["hash"]:
            log.debug(f"Already downloaded {local_path}")
            return
        else:
            log.warning(
                f"While updating, {local_path.name} local sha256 digest is "
                + f"{h.hexdigest()} but remote indicates it should be "
                + f"{entry['attachment']['hash']}, re-downloading."
            )

    log.debug(f"Downloading {url} to {local_path}")
    rsp = requests.get(url, stream=True)
    rsp.raise_for_status()

    local_path.write_bytes(rsp.content)

    h = hashlib.sha256()
    h.update(local_path.read_bytes())
    if h.hexdigest() != entry["attachment"]["hash"]:
        log.warning(
            f"While updating, {local_path.name} local sha256 digest is "
            + f"{h.hexdigest()} but remote indicates it should be "
            + f"{entry['attachment']['hash']}, raising exception."
        )
        raise ValueError(f"Hash mismatch on downloaded file {local_path}")


comment_re = re.compile(r"^\s*[#;]")


def parse_hosts_file(fd):
    host_strings = []
    for l in fd:
        line = l.strip()
        if line and not comment_re.match(line):
            host_strings.append(line)
    return host_strings


class IntermediatesDB(object):
    def __init__(self, *, db_path, download_pems=False):
        self.db_path = Path(db_path).expanduser()
        self.conn = sqlite3.connect(str(self.db_path / Path("intermediates.sqlite")))
        self.conn.row_factory = sqlite3.Row
        self.download_pems = download_pems
        self.intermediates_path = self.db_path / "intermediates"
        if self.download_pems:
            self.intermediates_path.mkdir(exist_ok=True)

        self.conn.execute(
            """
            CREATE TABLE IF NOT EXISTS intermediates (
                id TEXT PRIMARY KEY, last_modified TEXT, subject TEXT,
                subjectDN BLOB, derHash BLOB, pubKeyHash BLOB,
                crlite_enrolled BOOLEAN, whitelist BOOLEAN)"""
        )

    def __len__(self):
        with self.conn as c:
            cur = c.cursor()
            cur.execute("SELECT COUNT(*) FROM intermediates;")
            return cur.fetchone()[0]

    def __str__(self):
        return f"{len(self)} Intermediates"

    def update(self, *, collection_url, attachments_base_url):
        rsp = requests.get(collection_url)

        all_remote_ids = {x["id"] for x in rsp.json()["data"]}

        with self.conn as c:
            c.executemany(
                """INSERT INTO intermediates (id, last_modified,
                                    subject, subjectDN, derHash, pubKeyHash,
                                    crlite_enrolled, whitelist)
                             VALUES(:id, :last_modified, :subject, :subjectDN,
                                    :derHash, :pubKeyHash, :crlite_enrolled,
                                    :whitelist)
                             ON CONFLICT(id)
                             DO UPDATE SET id=:id, last_modified=:last_modified,
                                    subject=:subject, subjectDN=:subjectDN,
                                    derHash=:derHash, pubKeyHash=:pubKeyHash,
                                    crlite_enrolled=:crlite_enrolled,
                                    whitelist=:whitelist;
                            """,
                rsp.json()["data"],
            )

        if self.download_pems:
            log.info(f"Intermediates Update: Syncing intermediate certificates.")
            count = 0
            for entry in progressbar.progressbar(rsp.json()["data"]):
                local_path = self.intermediates_path / entry["id"]
                ensure_local(
                    base_url=attachments_base_url, entry=entry, local_path=local_path
                )
                count += 1

            log.info(f"Intermediates Update: {count} intermediates up-to-date.")

        with self.conn as c:
            cur = c.cursor()
            all_local_ids = {
                x["id"] for x in cur.execute("SELECT id FROM intermediates;").fetchall()
            }

            ids_to_remove = all_local_ids - all_remote_ids
            if ids_to_remove:
                log.info(
                    f"A total of {len(ids_to_remove)} intermediates have been removed. "
                    + "Syncing deletions."
                )

            for id_to_remove in ids_to_remove:
                path = self.intermediates_path / id_to_remove
                log.debug(f"Removing {path.name} (on disk={path.exists()})")
                if path.exists():
                    path.unlink()
                c.execute(
                    "DELETE FROM intermediates WHERE id=:id;", {"id": id_to_remove}
                )

    def issuer_by_DN(self, distinguishedName):
        with self.conn as c:
            cur = c.cursor()
            cur.execute(
                "SELECT id, subject, pubKeyHash, crlite_enrolled FROM intermediates "
                + "WHERE subjectDN=:dn LIMIT 1;",
                {"dn": base64.b64encode(bytes(distinguishedName)).decode("utf-8")},
            )
            row = cur.fetchone()
            if not row:
                return None
            data = {
                "subject": row["subject"],
                "spki_hash_bytes": base64.b64decode(row["pubKeyHash"]),
                "crlite_enrolled": row["crlite_enrolled"] == 1,
                "issuerId": IssuerId(base64.b64decode(row["pubKeyHash"])),
            }
            pem_path = Path(self.intermediates_path) / Path(row["id"])
            if pem_path.is_file():
                data["path"] = pem_path

            return data


class CRLiteDB(object):
    def __init__(self, *, db_path):
        self.db_path = Path(db_path).expanduser()
        self.filter_file = None
        self.stash_files = list()
        self.issuer_to_revocations = collections.defaultdict(list)
        self.filtercascade = None

        if self.db_path.is_dir():
            self.__load()
        else:
            self.db_path.mkdir()

    def __str__(self):
        count_revocations = sum(
            map(
                lambda x: len(self.issuer_to_revocations[x]), self.issuer_to_revocations
            )
        )
        return (
            f"Current filter: {self.filter_file.stem} with {self.filtercascade.layerCount()} "
            + f"layers and {self.filtercascade.bitCount()} bit-count, {len(self.stash_files)} "
            + f"stash files with {count_revocations} stashed revocations, up-to-date as of "
            + f"{self.latest_covered_date()} ({self.age()} ago)."
        )

    def filter_date(self):
        if not self.filter_file:
            return None
        time_str = self.filter_file.name.replace("Z-full", "")
        try:
            return datetime.fromisoformat(time_str).replace(tzinfo=timezone.utc)
        except ValueError as ve:
            log.warning(
                f"Couldn't decode filter path into timestamp, assuming it's right now: {ve}"
            )
            return datetime.now(tz=timezone.utc)

    def latest_stash_date(self):
        if not self.stash_files:
            return None
        time_str = self.stash_files[-1].name.replace("Z-diff", "")
        try:
            return datetime.fromisoformat(time_str).replace(tzinfo=timezone.utc)
        except ValueError as ve:
            log.warning(
                f"Couldn't decode stash path into timestamp, assuming it's right now: {ve}"
            )
            return datetime.now(tz=timezone.utc)

    def latest_covered_date(self):
        if self.stash_files:
            return self.latest_stash_date()
        return self.filter_date()

    def age(self):
        return datetime.now(tz=timezone.utc) - self.latest_covered_date()

    def load_filter(self, *, path):
        self.filter_file = path
        self.filtercascade = FilterCascade.from_buf(self.filter_file.read_bytes())
        self.issuer_to_revocations = collections.defaultdict(list)
        self.stash_files = list()

    def load_stashes(self, *, stashes):
        filter_date_str = self.filter_file.stem

        self.stash_files = list(
            filter(lambda x: str(x.name) > filter_date_str, sorted(stashes))
        )
        for path in self.stash_files:
            with path.open("rb") as f:
                for entry in readFromAdditionsList(f):
                    self.issuer_to_revocations[entry["issuerId"]].extend(
                        entry["revocations"]
                    )

    def __load(self):
        filters = sorted(self.db_path.glob("*-full"))
        if not filters:
            return

        self.load_filter(path=filters.pop())
        self.load_stashes(stashes=self.db_path.glob("*-diff"))

    def cleanup(self):
        filters = sorted(self.db_path.glob("*-full"))
        if filters:
            filters.pop()  # Leave the most recent
        for old_filter in filters:
            log.debug(f"Cleaning up old filter {old_filter}")
            old_filter.unlink()

        old_stashes = set(self.db_path.glob("*-diff")) - set(self.stash_files)
        for old_stash in old_stashes:
            log.debug(f"Cleaning up old stash {old_stash}")
            old_stash.unlink()

    def update(self, *, collection_url, attachments_base_url):
        rsp = requests.get(collection_url)
        entries = rsp.json()["data"]
        filter_entries = list(filter(lambda x: x["incremental"] is False, entries))
        stash_entries = sorted(
            filter(lambda x: x["incremental"] is True, entries),
            key=lambda x: x["details"]["name"],
        )

        if len(filter_entries) != 1:
            log.warning(
                f"Unexpected: Found more than one full filter at Remote Settings, "
                + f"found {len(filter_entries)}: {filter_entries}"
            )

        log.info(f"CRLite Update: Syncing CRLite filters.")

        all_entries = stash_entries + [filter_entries.pop()]

        for entry in progressbar.progressbar(all_entries):
            log.debug(f"Downloading Kinto entry={entry})")
            local_path = self.db_path / entry["details"]["name"]
            ensure_local(
                base_url=attachments_base_url, local_path=local_path, entry=entry
            )
            if entry["incremental"]:
                self.stash_files.append(local_path)
            else:
                self.filter_file = local_path

        self.__load()

    def download_to_db(self, *, base_url, entry):
        local_path = self.db_path / entry["details"]["name"]
        ensure_local(base_url=base_url, local_path=local_path, entry=entry)
        return local_path

    def validity_window_status(self, issue_time, expire_time):
        return "Too New"

    def revocation_status(self, certId):
        results = {}

        revoked_in_crlite = certId.to_bytes() in self.filtercascade
        if revoked_in_crlite:
            results["via_filter"] = self.filter_file.name

        revoked_in_stash = False
        for path in self.stash_files:
            with path.open("rb") as f:
                for entry in readFromAdditionsList(f):
                    if (
                        entry["issuerId"] == certId.issuerId
                        and certId in entry["revocations"]
                    ):
                        revoked_in_stash = True
                        results["via_stash"] = path.name
                        break

        results["revoked"] = revoked_in_crlite or revoked_in_stash
        return results


# Theoretically this should get us the raw serial, but it... doesn't.
# class RawSerial(univ.OctetString):
#     tagSet = tag.initTagSet(
#         tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x02)
#     )
#     subtypeSpec = constraint.ConstraintsIntersection()
#     typeId = univ.OctetString.getTypeId()


class RawTBSCertificate(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.DefaultedNamedType(
            "version",
            rfc2459.Version("v1").subtype(
                explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            ),
        ),
        namedtype.NamedType("serialNumber", univ.Integer()),
        namedtype.NamedType("signature", rfc2459.AlgorithmIdentifier()),
        namedtype.NamedType("issuer", univ.Any()),
        namedtype.NamedType("validity", rfc2459.Validity()),
        namedtype.NamedType("subject", rfc2459.Name()),
        namedtype.NamedType("subjectPublicKeyInfo", rfc2459.SubjectPublicKeyInfo()),
        namedtype.OptionalNamedType(
            "issuerUniqueID",
            rfc2459.UniqueIdentifier().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            ),
        ),
        namedtype.OptionalNamedType(
            "subjectUniqueID",
            rfc2459.UniqueIdentifier().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 2)
            ),
        ),
        namedtype.OptionalNamedType(
            "extensions",
            rfc2459.Extensions().subtype(
                explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 3)
            ),
        ),
    )


class RawCertificate(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType("tbsCertificate", RawTBSCertificate()),
        namedtype.NamedType("signatureAlgorithm", rfc2459.AlgorithmIdentifier()),
        namedtype.NamedType("signatureValue", univ.BitString()),
    )


class CRLiteQueryResult(object):
    def __init__(self, *, name, issuer, cert_id, crlite_db, not_before, not_after):
        self.name = name
        self.issuer = issuer
        self.cert_id = cert_id
        self.via_stash = None
        self.via_filter = None

        if not issuer:
            self.state = "Unknown Issuer"
            return

        if not issuer["crlite_enrolled"]:
            self.state = "Not Enrolled"
            return

        if crlite_db.latest_covered_date() < not_before:
            self.state = "Too New"
            return

        if crlite_db.latest_covered_date() > not_after:
            self.state = "Expired"
            return

        rev_status = crlite_db.revocation_status(self.cert_id)

        if rev_status["revoked"] is True:
            if "via_stash" in rev_status:
                self.via_stash = rev_status["via_stash"]
            if "via_filter" in rev_status:
                self.via_filter = rev_status["via_filter"]
            self.state = "Revoked"
            return

        self.state = "Valid"

    def __str__(self):
        return f"{self.name} {self.cert_id} state={self.state}"

    def is_revoked(self):
        return self.state == "Revoked"

    def result_icon(self):
        if self.state == "Revoked":
            return "⛔️"
        if self.state == "Not Enrolled":
            return "❔"
        if self.state == "Valid":
            return "👍"
        if self.state == "Too New":
            return "🐇"
        if self.state == "Expired":
            return "⏰"
        if self.state == "Unknown Issuer":
            return "⁉️"
        return ""

    def print_query_result(self):
        padded_name = self.name + " " * 5
        padding = "".ljust(len(padded_name))

        if not self.issuer:
            self.state = "Unknown Issuer"
        else:
            enrolled_icon = "✅" if self.issuer["crlite_enrolled"] else "❌"

            print(f"{padded_name} Issuer: {self.issuer['subject']}")
            print(f"{padding} Enrolled in CRLite: {enrolled_icon}")
            print(f"{padding} {self.cert_id}")
            if self.via_filter:
                print(f"{padding} Revoked via CRLite filter: {self.via_filter}")
            if self.via_stash:
                print(f"{padding} Revoked via Stash: {self.via_stash}")

        print(
            f"{padding} Result: {self.result_icon()} {self.state} {self.result_icon()}"
        )

    def log_query_result(self):
        if not self.issuer:
            log.warning(f"{self.name} Unknown issuer")
            return

        logdata = {
            "name": self.name,
            "issuer": self.issuer["subject"],
            "state": self.state,
            "state_icon": self.result_icon(),
            "enrolled_in_crlite": "✅" if self.issuer["crlite_enrolled"] else "❌",
            "cert_id": self.cert_id,
            "via_filter": self.via_filter,
            "via_stash": self.via_stash,
        }

        if self.state == "Revoked":
            log.warning(logdata)
        else:
            log.info(logdata)


class CRLiteQuery(object):
    def __init__(self, *, intermediates_db, crlite_db):
        self.intermediates_db = intermediates_db
        self.crlite_db = crlite_db
        self.context = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS, ca_certs=None)

    def gen_from_host(self, host, port):
        try:
            with self.context.wrap_socket(
                socket.socket(socket.AF_INET),
                server_hostname=host,
                do_handshake_on_connect=True,
            ) as conn:
                conn.connect((host, port))
                yield conn.getpeercert(binary_form=True)
        except ssl.SSLError as se:
            logging.warning(f"Failed to fetch from {host}:{port}: {se}")
        except TimeoutError:
            logging.warning(f"Failed to fetch from {host}:{port}: timed out")
        except Exception as e:
            logging.warning(f"Failed to fetch from {host}:{port}: {e}")
        return

    def gen_from_pem(self, file_obj):
        while True:
            data = pem.readPemFromFile(file_obj)
            if not data:
                return
            yield data

    def query(self, *, name, generator):
        for data in generator:
            cert, rest = der_decoder(data, asn1Spec=RawCertificate())
            assert not rest, f"unexpected leftovers in ASN.1 decoding: {rest}"

            serial_number = cert.getComponentByName(
                "tbsCertificate"
            ).getComponentByName("serialNumber")

            # Serial numbers are represented big-endian
            serial_bytes = int(serial_number).to_bytes(
                (int(serial_number).bit_length() + 7) // 8,
                byteorder="big",
                signed=False,
            )

            issuerDN = cert.getComponentByName("tbsCertificate").getComponentByName(
                "issuer"
            )
            issuer = self.intermediates_db.issuer_by_DN(issuerDN)
            if not issuer:
                yield CRLiteQueryResult(
                    name=name,
                    issuer=None,
                    cert_id=None,
                    crlite_db=self.crlite_db,
                    not_before=None,
                    not_after=None,
                )
                continue

            cert_id = CertId(issuer["issuerId"], serial_bytes)

            validity = cert.getComponentByName("tbsCertificate").getComponentByName(
                "validity"
            )
            not_before = validity.getComponentByName("notBefore").getComponent()
            not_before = datetime.strptime(str(not_before), "%y%m%d%H%M%SZ").replace(
                tzinfo=timezone.utc
            )
            not_after = validity.getComponentByName("notAfter").getComponent()
            not_after = datetime.strptime(str(not_after), "%y%m%d%H%M%SZ").replace(
                tzinfo=timezone.utc
            )

            yield CRLiteQueryResult(
                name=name,
                issuer=issuer,
                cert_id=cert_id,
                crlite_db=self.crlite_db,
                not_before=not_before,
                not_after=not_after,
            )



from setuptools import setup


setup(
    name="moz_crlite_query",
    version="0.3.4",
    description="Query CRLite for a certificate, or certificate information",
    long_description="Use this tool to download and maintain CRLite information from "
    + "Mozilla's Remote Settings infrastructure, and query it.",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3",
    ],
    keywords="bloom filter cascade multi level mlbf crlite",
    url="http://github.com/mozilla/moz_crlite_query",
    author="J.C. Jones",
    author_email="jc@mozilla.com",
    license="Mozilla Public License 2.0 (MPL 2.0)",
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "deprecated",
        "filtercascade",
        "glog",
        "moz-crlite-lib",
        "progressbar2",
        "pyasn1-modules",
        "requests",
    ],
    packages=["crlite_query"],
    entry_points={
        "console_scripts": ["moz_crlite_query=crlite_query.query_cli:main"],
    },
)
